/*
 *  libbab.cc -- BabyTrans ( Babylon Translator front-end for GTK )
 *
 *  Copyright (C) 1999  Frederic Jolliton -- <fjolliton@fnac.net>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */
#include <stdio.h>
#include <string.h>
#include <ctype.h>

#define MAX_WORD_LENGTH 63
#define BLOCK_SIZE 1024

#define _(a) a

#define warn(s) (fprintf(stderr,"%s:%d: %s\n",__FILE__,__LINE__,(s)))

#ifdef DEBUG
# define print_var(fmt,v) (fprintf(stderr,"%s:%d: %s=" fmt "\n",\
				    __FILE__,__LINE__,\
				    #v,v))
#else
#define print_var(fmt,v)
#endif

/* System dependant definitions ! */
typedef unsigned char   u8 ;
typedef unsigned short  u16 ;
typedef unsigned long   u32 ;

/* ==[ TYPES ]============================================================ */

/* --[ item::type ]------------------------------------------------------- */
struct item_ {
   char *word;
   char *attribute;
   char *definition;
   struct item_ *next;
};
typedef struct item_ item_t;


/* --[ babylon::type ]---------------------------------------------------- */
#define BABYLON_STRUCT
struct babylon_struct{
   FILE *file_idx;
   FILE *file_def;
   int ok; /*bool: is ready to translate*/
   item_t *definition;
   item_t *curr; /* to itere between definitions */
};
#include "babylon.h"

/* ==[ PROTOTYPES OF PRIVATE FUNCTIONS ]================================== */
/* Make a new definition node */
/* pos indicate the position of the definition in file_def file */
void add_definition(babylon_t *bab, size_t pos );
/* reads the word at current position, and return it */
char *read_word( babylon_t *bab , unsigned int length );
/* trim a word, convert it to lower case, and put '_' at the end if needed 
 * to force it to have a length superior to 3 */
char *babylon_trim( const char *word );
/* read a value on #n bytes */
inline unsigned long read_val( FILE *f , int n );
/* Compute an index from word's first 3 characters */
long word_index( const char * word );
/* add a new definition node to the bab object */
/* pos indicate the position of the definition in file_def file */
void add_definition(babylon_t *bab, size_t pos );

/* ==[ CODE ]============================================================= */

/* --[ item::add ]-------------------------------------------------------- */
item_t *item_add(char *word,char *attribute,char *definition,item_t *queue){
   item_t *res;
   if (!(res=(item_t*)malloc(sizeof(item_t)))){
      warn("Allocation problem");
      return queue;
   }
   res->word=strdup(word);
   res->attribute=strdup(attribute);
   res->definition=strdup(definition);
   res->next=queue;
   return res;
}

/* --[ item::free ]------------------------------------------------------- */ 
void item_free(item_t *i){
   item_t *j;
   
   while (i != NULL){
      j=i->next;
      free(i->word);
      free(i->attribute);
      free(i->definition);
      free(i);
      i=j;
   }
}
/* --[ item::concat ]----------------------------------------------------- */ 
item_t *item_concat(item_t *a,item_t *b){
   item_t *c;
   
   if (a==NULL){
      return b;
   }
   if (b==NULL){
      return a;
   }
   c=a;
   while (c->next != NULL)
      c=c->next;
   c->next=b;
   return a;
}


/* --[ babylon::creator ]------------------------------------------------- */
babylon_t *babylon_new(char *name_idx,char *name_def){
   babylon_t *bab;
   
   if (!(bab = (babylon_t *)malloc(sizeof(babylon_t)))) {
      warn("Allocation error\n");
      return NULL;
   }
   if (!(bab->file_idx=fopen(name_idx,"r"))){
      fprintf(stderr,"Can't open %s\n",name_idx);
      free(bab);
      return NULL;
   }
   if (!(bab->file_def=fopen(name_def,"r"))){
      fprintf(stderr,"Can't open %s\n",name_def);
      fclose(bab->file_idx);
      free(bab);
      return NULL;
   }
   bab->ok=1;
   bab->definition=NULL;
   return bab;
}

/* --[ babylon::destructor ]---------------------------------------------- */
void babylon_free(babylon_t *b){
   if (b!=NULL){
      item_free(b->definition);
      if (b->file_idx!=NULL)
	fclose(b->file_idx);
      if (b->file_def!=NULL)
	fclose(b->file_def);
      free(b);
   }
}

/* --[ babylon::translate ]----------------------------------------------- */
int babylon_translate(babylon_t *bab,const char *w) {
   char  *word=babylon_trim(w); /* word, ready to be translated */
   long   main_index ; /* main index, as computed by word_index() */
   u32    idx_idx , idx_def ; /* index in files bab->file_{idx|def} */
   int    wordCount ; /* number of words during second indirection */
   size_t byte_to_skip ; /* after second indirection */
   int    c ;
   char   ch ;
   int    i,j; /* loop counters */
   u16    nbWord ; /* number of word */
   u32    def_adr ; /* address of the definition in file */
   
   /* min/max size of words in this cell after the second indirection */
   size_t              min_size , max_size ;

   
   /* frees any previous definitions */
   item_free(bab->definition);
   bab->definition=NULL;
   
   /* 1) Compute main index (first indirection) */
   main_index = 100 + 4 * word_index( word ) ;

   /* 2) Seek to both file (apply the first indirection) */
   fseek( bab->file_idx , main_index , SEEK_SET ) ;
   fseek( bab->file_def , main_index , SEEK_SET ) ;
   
   /* 3) Read the second indirection */
   idx_idx = read_val( bab->file_idx , 4 ) ;
   idx_def = read_val( bab->file_def , 4 ) ;
   if ( idx_def == read_val( bab->file_def , 4 ) )
     return 1 ;
     /* if (curr index in def)=(next index in def) then
      *   entry not defined, return empty list */

   /* 4) Follow the second indirection and read cells */
   fseek( bab->file_idx , idx_idx , SEEK_SET ) ;

   min_size = (int)read_val( bab->file_idx , 1 ) ;
   max_size = (int)read_val( bab->file_idx , 1 ) ;

   if ( ( strlen(word) < min_size ) || ( strlen(word) > max_size ) )
     return 1 ;
     /* if not min_size <= length <= max_size, this word can't be found */
     /*   return empty list */

   /* 4.1) Read the cell corresponding to my actual length of word.
    *      It gives me the number of word I have for this length.
    *      I also count how many byte I will have to skip to reach
    *       the beginning of my words */
   
   wordCount = 0 ;
   byte_to_skip = 0 ;
   for ( i = min_size ; i < strlen(word) ; ++ i ) {
      nbWord = (int)read_val( bab->file_idx , 2 ) ;
      wordCount += nbWord ;
      byte_to_skip += ( i - 3 ) * nbWord ;
   }
   nbWord = (int)read_val( bab->file_idx , 2 ) ; /* this is my value */
   byte_to_skip += ( max_size - strlen(word) ) * 2 ;
   
   /* 4.2) Go to the begining of words of my length */
   fseek( bab->file_idx , byte_to_skip , SEEK_CUR ) ;

   /* 4.3) Search the indirection for my word:
    *      I compare the next chars with my word.
    *      if they are equal (j==length after comparaison):
    *         I read the final indirection in the def file.
    *         (this could be an link to another def, if so, follow it)
    *         make a definition of this indirection.
    *      if they differ, I skip this entry.
    */

   for ( i = 0 ; i < (int)nbWord ; ++ i ) {
      for ( j = 3 ;
	   j < strlen(word)
	   && word[ j ] == (ch=getc(bab->file_idx)) ;
	   ++ j )
	;/* the comparison */
      if ( j == strlen(word) ) { /* matching !! */
	 fseek( bab->file_def , idx_def + 4 * wordCount , SEEK_SET ) ;
	 def_adr = read_val( bab->file_def , 4 ) ;

	 if ( def_adr & 0xFF000000UL ) { /* link to another def reference */
	    main_index = 100 + 4 * ( ( def_adr >> 16 ) & 0xFFFF ) ;
	    fseek (bab->file_def, main_index , SEEK_SET) ;
	    idx_def = read_val( bab->file_def , 4 ) ;

	    fseek(bab->file_def, 
		  idx_def + 4 * ( def_adr & 0xFFFF ), 
		  SEEK_SET ) ;
	    def_adr = read_val( bab->file_def , 4 ) ;
	 }
	 /* read the def to build a new definition */
	 add_definition(bab, def_adr );
      } else {
	 /* not matched. Skip the rest of this word. But beware: */
	 /*   a character has been read without j incremented */
	 fseek(bab->file_idx, strlen(word) - j - 1 , SEEK_CUR ) ;
      }
      ++ wordCount ;
   }

   return 1 ;
}
/* --[ babylon::def::first ]--------------------------------------------- */
void babylon_def_first(babylon_t *bab) {
   bab->curr=bab->definition;
}
/* --[ babylon::def::next ]---------------------------------------------- */
void babylon_def_next(babylon_t *bab) {
   if (bab->curr != NULL) {
      bab->curr = bab->curr->next;
   }
}
/* --[ babylon::def::curr ]---------------------------------------------- */
void babylon_def_curr(babylon_t *bab,
		      char **word,char **attribute,char **definition) {
   if (bab->curr != NULL){
      *word=bab->curr->word;
      *attribute=bab->curr->attribute;
      *definition=bab->curr->definition;
   } else {
      *word = *attribute = *definition = NULL;
   }
}
/* --[ babylon::def::last ]---------------------------------------------- */
int  babylon_def_last(babylon_t *bab) {
   return (bab->curr == NULL);
}
/* --[ babylon::private functions ]-------------------------------------- */
/* read a value on #n bytes */
inline unsigned long read_val( FILE *f , int n ) {
   int i;
   unsigned long       r = 0 ;
   for ( i = 0 ; i < n ; ++ i )
     r |= ( (unsigned long)(unsigned char)getc(f) ) << 8*i ;
   return r ;
}

/* Compute an index from word's first 3 characters */
long word_index( const char * word ) {
    long                idx = 0 ;
    switch( strlen(word) ) {
    default:
        idx += word[ 2 ] - 'a' + 2 ;
    case 2:
        idx += ( word[ 1 ] - 'a' + 2 ) * 28 ;
    case 1:
        idx += ( word[ 0 ] - 'a' + 2 ) * 28 * 28;
    case 0:
        break ;
    }
    return idx ;
}

inline char bab_to_char( int c ) {
   return "abcdefghijklmnopqrstuvwxyz* **'*"[ c ] ;
}   

/* trim a word, convert it to lower case, and put '_' at the end if needed 
 * to force it to have a length superior to 3 */
char *babylon_trim( const char *word ){
   char *word2; /* a copy of the string to modifiate it */
   char *trimed_word; /* first non white in word2 */
   char *p; /* a pointer on string to iterate chars */
   int i; /* counter in loop */
   
   /* Trim the word */
   word2=strdup(word);
   trimed_word=word2;
   while (isblank(*trimed_word)) 
     trimed_word++;
   
   p=trimed_word;
   while (isalpha(*p))
     p++;
   *p='\0';
   if (trimed_word==p){
      fprintf(stderr, "Invalid word");
      return 0;
   }
   
   if  ( strlen(trimed_word) >= MAX_WORD_LENGTH ) {
      fprintf(stderr, "Word is tool long" ) ;
      return 0 ;
   }

   for ( i = 0 ; i < strlen(trimed_word) ; ++ i )
        trimed_word[ i ] = tolower( trimed_word[ i ] ) ;

   while ( strlen(trimed_word) < 3 ) {
      trimed_word[strlen(trimed_word)+1] = '\0';
      trimed_word[strlen(trimed_word)] = '_' ;
   }
   
   p=strdup(trimed_word);
   free(word2);
   return p;
}

/* Make a new definition node */
/* pos indicate the position of the definition in file_def file */
void add_definition(babylon_t *bab, size_t pos )
{
   item_t def ;
   char *word;
   char *attribute,*human_attrib;
   char *definition,*d;
   char c1,c2; /* to add a space after comma in def */
   int size; /* size to read */
   int i,j; /* loop counters */
   int v; /* to read vals */
   /* to read definition */
   int mask = 0 ;
   int c = 0 ;
   int prev_c ;

   
   if ((!(attribute=(char *)malloc(17*sizeof(char)))) ||
       (!(definition=(char *)malloc(BLOCK_SIZE)))) {
      fprintf(stderr,"Allocation problem");
      return;
   }

   fseek(bab->file_def, pos , SEEK_SET) ;

   /* reads the attribute */
   d=attribute;
   for ( i = 0 ; i < 7 ; ++ i ) {
      v = (int)read_val( bab->file_def , 1 ) ;
      *(d++) = "0123456789ABCDEF"[ v/16 ] ;
      *(d++) = "0123456789ABCDEF"[ v%16 ] ;
   }
   attribute[16]='\0';
   
   /* convert the attribute to a human readable form */
   //   print_var("%c",attribute[2]); print_var("%c",attribute[3]);
   if ( attribute[ 2 ] > '1' || attribute[ 3 ] != '0' ) {         /* noun */
     human_attrib = _( "n." ) ;
   } else if ( attribute[ 6 ] == '1' ) {                          /* adjective */
     human_attrib = _( "a." ) ;
   } else if ( attribute[ 6 ] != '0' || attribute[ 9 ] != '0' ) { /* verb */
     human_attrib = _( "v." ) ;
   } else if ( attribute[ 7 ] != '0' ) {                          /* adverb */
     human_attrib = _( "a." ) ;
   } else if ( attribute[ 8 ] == '1' ) {                          /* adverb */
     human_attrib = _( "adv." ) ;
   } else 
     human_attrib = attribute;
   attribute[9]='\0';

   /* reads the word */
   size = (int)read_val( bab->file_def , 1 ) ;

   word = read_word( bab , size ) ;

   /* Reads the definition.
    *  The difficulty commes from:
    *    1) escape codes which are used to commpress '...' and others strings.
    *       They are coded on two chars 
    *    2) Masks used to write chars (encryption ?) */
   size = (int)read_val( bab->file_def , 1 ) ;
#ifdef DEBUG
   if ( size == 0 )
     warn( "Null definition" ) ;
#endif
   d=definition;
   while ( size > 0 ) {
      prev_c = c ;
      c = (int)read_val( bab->file_def , 1 ) ^ mask ;
      mask ^= 0x80 ;
      if ( c == 0 )
	continue ;
      if ( prev_c == 0x7D ) {
	 if ( c == 0XEF ) {
	    *(d++) = ',';
	    *(d++) = ' ';
	    *(d++) = ' ';
	    size -= 3 ;
	 } else {
	    *(d++) = prev_c ;
	    *(d++) = c ;
	    size -= 2 ;
	 }
      } else if ( prev_c == 0x7F ) { // [ 7F EF ] => ". "
	 if ( c == 0xEF ) {
	    *(d++) = '.';
	    *(d++) = ' '; 
	    size -= 3 ; // Considered as 3 chars.. Hum..
	 }
      } else if ( prev_c == 0xFF ) {
	 if ( c == 0xF3 ) { // [FF F3]  => "..;"
	    *(d++) = '.';
	    *(d++) = '.'; 
	    *(d++) = ';'; 
	 } else if ( c == 0xEF ) { // [FF EF] => ".. "
	    *(d++) = '.';
	    *(d++) = '.'; 
	    *(d++) = ' ';
	 } else if ( c == 0xFF ) { // [FF FF] => "..."
	    *(d++) = '.';
	    *(d++) = '.'; 
	    *(d++) = '.';
	 }
	 size -= 3 ;
	 c = 0 ; // 'clear previous char'
      } else if ( prev_c == 0xFB ) {
	 if ( c == 0xFF ) { // add ".."
	    *(d++) = '.';
	    *(d++) = '.'; 
	    size -= 3 ; // Considered as 3 chars.. Hum..
	    c = 0 ; // 'clear previous char'
	 } else { /* waited for nothing: not an escape code */
	    *(d++) = prev_c ;
	    *(d++) = c;
	    size -= 2 ;
	 }
	 continue ;
      } else if ( c == 0xFB || c == 0xFF || c == 0x7D || c == 0x7F ) {
	 ; /* wait */
      } else { /* normal case. Grab the char to the def */
	 *(d++) = c ;
	 -- size ;
      }
   }
   *d='\0';

   /* Do s/;/\n/ in definition */
   for ( i = 0 ; i < strlen(definition) ; ++ i ) 
     if ( definition[ i ] == ';' ) 
       definition[ i ] = '\n' ; 

   /* Put an espace after comma in definition if needed */
   size=strlen(definition);
   for ( i = 0 ; i < size ; ++ i ) {
     if ( ( definition[ i ] == ',' ) &&
	  ( definition[ i + 1 ] != ' ' ) ) {
       c1 = definition[ i+1 ];
       definition[ i+1 ]=' ';
       for ( j=i+1 ; j < size+1 ; ++ j) {
	 c2=definition[j+1];
	 definition[j+1]=c1;
	 c1=c2;
       }
       definition[++size]='\0';
     }
   }

   bab->definition=item_add(word,human_attrib,definition,bab->definition);
   free(word);
   free(attribute);
   free(definition);
}

/* reads the word at current position, and return it */
char *read_word( babylon_t *bab , unsigned int length )
{
   static char*        compact_table[] = {
      "<0>" , "ion" , "ies" , "ing" , "ous" , "ses" ,        /*  0 ..  5 */
      "al" , "an" ,                                          /*  6 ..  7 */
      "at" , "ed" , "en" , "er" , "es" , "ia" , "ic" , "ie", /*  8 .. 15 */
      "in" , "io" , "is" , "it" , "le" , "ly" , "ne" , "on", /* 16 .. 23 */
      "or" , "ra" , "se" , "ss" , "st" , "te" , "ti" , "th"  /* 24 .. 31 */
   } ;

   char *definition,*d;
   u16 data;
   int lsb,msb;

   if (!(definition=(char*)malloc((length+1)*sizeof(char)))) {
      warn("Allocation problem");
      return NULL;
   }

   d=definition;
   while ( length > 0 ) {
      data = read_val( bab->file_def , 2 ) ;
      if ( data < 32768 ) {
	 *(d++) = bab_to_char( (int)( data & 0x1F ) ) ;
	 data >>= 5 ;
	 *(d++) = bab_to_char( (int)( data & 0x1F ) ) ;
	 data >>= 5 ;
	 *(d++) = bab_to_char( (int)( data & 0x1F ) ) ;
	 length -= 3 ;
      } else {
	 lsb = (int)( data & 0x7F ) ;
	 msb = (int)( ( data >> 8 ) & 0x7F ) ;

	 if ( lsb >= 32 ) {
	    *(d++) = (char)lsb ;
	    -- length ;
	 } else if ( lsb >= 6 ) {
	    *(d++) = compact_table[ lsb ][0] ;
	    *(d++) = compact_table[ lsb ][1] ;
	    length -= 2 ;
	 } else {
	    *(d++) = compact_table[ lsb ][0] ;
	    *(d++) = compact_table[ lsb ][1] ;
	    *(d++) = compact_table[ lsb ][2] ;
	    length -= 3 ;
	 }
	 if ( length == 0 )
	   break ;
	 if ( length < 0 ) {
	    //              warn( "buffer underflow" ) ;
	    return definition ;
	 }

	 if ( msb >= 32 ) {
	    *(d++) = (char)msb ;
	    -- length ;
	 } else if ( msb >= 6 ) {
	    *(d++) = compact_table[ msb ][0] ;
	    *(d++) = compact_table[ msb ][1] ;
	    length -= 2 ;
	 } else {
	    *(d++) = compact_table[ msb ][0] ;
	    *(d++) = compact_table[ msb ][1] ;
	    *(d++) = compact_table[ msb ][2] ;
	    length -= 3 ;
	 }
      }
   }
   *d = '\0';
   return definition ;
}

void print_trans(babylon_t *bab , char * w) {
   char *word,*attribute,*definition;

   babylon_translate(bab,w);
   babylon_def_first(bab);
   while (!babylon_def_last(bab)) {
      babylon_def_curr(bab,&word,&attribute,&definition);
      printf("__%s (%s)__\n%s\n\n",word,attribute,definition);
      babylon_def_next(bab); 
   }
}

main (int argc, char *argv[]){
  babylon_t *bab;
  int i;

  if (argc < 3) {
    fprintf (stderr,"%s: Usage\n%s <english_dic> <trad_dic> word [word..]\n\
example:\n%s /usr/local/lib/babytrans/english.dic /usr/local/lib/babytrans/EngtoFre.dic cat\n",
	     argv[0],argv[0],argv[0]);
    exit (1);
  }
  bab=babylon_new(argv[1],argv[2]);
  for (i=3; i<argc;i++){
    print_trans(bab , argv[i]);
  }
}
